import base64
import html
import json

import requests

from biliup.config import config
from biliup.plugins.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import match1, logger


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m)\.)?huya\.com')
class Huya(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.huya_danmaku = config.get('huya_danmaku', False)

    def check_stream(self):
        logger.debug(self.fname)
        try:
            res = requests.get(self.url, timeout=5, headers=self.fake_headers)
            res.close()
        except:
            logger.warning("虎牙：" + self.url + "：获取res错误")
            return False
        huya = None
        if match1(res.text, '"stream": "([a-zA-Z0-9+=/]+)"'):
            huya = base64.b64decode(match1(res.text, '"stream": "([a-zA-Z0-9+=/]+)"')).decode()
        elif match1(res.text, 'stream: ([\w\W]+)'):
            huya = res.text.split('stream: ')[1].split('};')[0].strip()
            if json.loads(huya)['vMultiStreamInfo']:
                huya = res.text.split('stream: ')[1].split('};')[0].strip()
            else:
                huya = None
        if huya:
            try:
                huyacdn = config.get('huyacdn', 'AL')
                huyajson1 = json.loads(huya)['data'][0]['gameStreamInfoList']
                huyajson2 = json.loads(huya)['vMultiStreamInfo']
                ratio = huyajson2[0]['iBitRate']
                ibitrate_list = []
                sdisplayname_list = []
                for key in huyajson2:
                    ibitrate_list.append(key['iBitRate'])
                    sdisplayname_list.append(key['sDisplayName'])
                    if len(sdisplayname_list) > len(set(sdisplayname_list)):
                        ratio = max(ibitrate_list)
                huyajson = huyajson1[0]
                for cdn in huyajson1:
                    if cdn['sCdnType'] == huyacdn:
                        huyajson = cdn
                absurl = f'{huyajson["sFlvUrl"]}/{huyajson["sStreamName"]}.{huyajson["sFlvUrlSuffix"]}?' \
                        f'{huyajson["sFlvAntiCode"]}'
                self.raw_stream_url = html.unescape(absurl) + "&ratio=" + str(ratio)
                self.room_title = json.loads(huya)['data'][0]['gameLiveInfo']['introduction']
                return True
            except:
                logger.warning("虎牙：" + self.url + "：json解析错误")
                return False

    async def danmaku_download_start(self, filename):
        if self.huya_danmaku:
            logger.info("开始弹幕录制")
            self.danmaku = DanmakuClient(self.url, filename + "." + self.suffix)
            await self.danmaku.start()

    def close(self):
        if self.huya_danmaku:
            self.danmaku.stop()
            logger.info("结束弹幕录制")


import requests
from ykdl.util.match import match1

from biliup.config import config
from biliup.plugins.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase
from ..plugins import logger

# import http.client
# http.client.HTTPConnection._http_vsn = 10
# http.client.HTTPConnection._http_vsn_str = 'HTTP/1.0'


@Plugin.download(regexp=r'(?:https?://)?(?:(?:www|m)\.)?douyu\.com')
class Douyu(DownloadBase):
    def __init__(self, fname, url, suffix='flv'):
        super().__init__(fname, url, suffix)
        self.douyu_danmaku = config.get('douyu_danmaku', False)

    def check_stream(self):
        if len(self.url.split("douyu.com/")) < 2:
            logger.error("斗鱼：" + self.url + "：地址错误")
            return False
        
        try:
            html = requests.get(self.url).text
            vid = match1(html, r'\$ROOM\.room_id\s*=\s*(\d+)',
                        r'room_id\s*=\s*(\d+)',
                        r'"room_id.?":(\d+)',
                        r'data-onlineid=(\d+)')
            if not vid:
                logger.debug("斗鱼：" + self.url + "：被关闭或不存在")
                return False
        except:
            logger.warning("斗鱼：" + self.url + "：获取vid错误")
            return False
        
        try:
            roominfo = requests.get(f"https://www.douyu.com/betard/{vid}").json()['room']
            if roominfo['show_status'] != 1 or roominfo['videoLoop'] != 0:
                logger.debug("斗鱼：" + vid + "：未开播或正在放录播")
                return False
            self.room_title = roominfo['room_name']
            html_h5enc = requests.get(f'https://www.douyu.com/swf_api/homeH5Enc?rids={vid}').json()
            js_enc = html_h5enc['data']['room' + vid]
            params = {
                'cdn': config.get('douyucdn', 'tct-h5'),
                'iar': 0,
                'ive': 0,
                'rate': 0,
            }
        except:
            logger.warning("斗鱼：" + vid + "：获取roominfo错误")
            return False
        
        try:
            ub98484234(js_enc, vid, params)
        except:
            logger.error('请至少安装一个 Javascript 解释器')
            return False
        live_data = get_play_info(vid, self.fake_headers, params)
        if type(live_data) is not dict:
            return False
        self.raw_stream_url = f"{live_data.get('rtmp_url')}/{live_data.get('rtmp_live')}"
        return True

    async def danmaku_download_start(self, filename):
        if self.douyu_danmaku:
            logger.info("开始弹幕录制")
            self.danmaku = DanmakuClient(self.url, filename + "." + self.suffix)
            await self.danmaku.start()

    def close(self):
        if self.douyu_danmaku:
            self.danmaku.stop()
            logger.info("结束弹幕录制")

def get_play_info(vid, headers, params):
    try:
        html_content = requests.post(f'https://www.douyu.com/lapi/live/getH5Play/{vid}', headers=headers, params=params).json()
        live_data = html_content["data"]
    except:
        logger.warning(f"douyu：get_play_info：请求出错：https://www.douyu.com/lapi/live/getH5Play/{vid}")
        return None
    # 禁用斗鱼主线路
    if not type(live_data) == type({}):
        return live_data
    if not live_data['rtmp_cdn'].endswith('h5') or 'akm' in live_data['rtmp_cdn']:
        params['cdn'] = 'tct-h5'
        return get_play_info(vid, headers, params)
    return live_data

def ub98484234(js_enc, vid, params):
    import jsengine
    import uuid
    import time
    from urllib.parse import parse_qs
    did = uuid.uuid4().hex
    tt = str(int(time.time()))
    crypto_js = requests.get('https://cdn.staticfile.org/crypto-js/4.1.1/crypto-js.min.js').text
    js = jsengine.JSEngine(js_enc + crypto_js)
    query = js.call('ub98484234', vid, did, tt)
    params.update({k: v[0] for k, v in parse_qs(query).items()})


import asyncio
import logging
import os
import re
import subprocess
import sys
import threading
import time
from urllib.parse import urlparse

import stream_gears

from biliup.config import config

logger = logging.getLogger('biliup')


class DownloadBase:
    def __init__(self, fname, url, suffix=None, opt_args=None):
        self.danmaku = None
        self.room_title = None
        if opt_args is None:
            opt_args = []
            # 主播单独传参覆盖全局设置。例如新增了一个全局的filename_prefix参数，在下面添加self.filename_prefix = config.get('filename_prefix'),
            # 即可通过self.filename_prefix在下载或者上传时候传递主播单独的设置参数用于调用（如果该主播有设置单独参数，将会优先使用单独参数；如无，则会优先你用全局参数。）
        self.fname = fname
        self.url = url
        self.suffix = suffix
        self.title = None
        self.live_cover_path = None
        self.downloader = config.get('downloader', 'stream-gears')
        # ffmpeg.exe -i  http://vfile1.grtn.cn/2018/1542/0254/3368/154202543368.ssm/154202543368.m3u8
        # -c copy -bsf:a aac_adtstoasc -movflags +faststart output.mp4
        self.raw_stream_url = None
        self.filename_prefix = config.get('filename_prefix')
        self.use_live_cover = config.get('use_live_cover')
        self.opt_args = opt_args
        self.fake_headers = {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'zh-CN,zh;q=0.8,en-US;q=0.5,en;q=0.3',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.159 Safari/537.36'
        }

        self.default_output_args = [
            '-bsf:a', 'aac_adtstoasc',
        ]
        if config.get('segment_time'):
            self.default_output_args += \
                ['-segment_time', f"{config.get('segment_time', '00:50:00')}"]
        else:
            self.default_output_args += \
                ['-fs', f"{config.get('file_size', '2621440000')}"]

    def check_stream(self):
        logger.debug(self.fname)
        raise NotImplementedError()

    def download(self, filename):
        if self.filename_prefix:  # 判断是否存在自定义录播命名设置
            filename = (self.filename_prefix.format(streamer=self.fname, title=self.room_title).encode(
                'unicode-escape').decode()).encode().decode("unicode-escape")
        else:
            filename = f'{self.fname}%Y-%m-%dT%H_%M_%S'
        filename = get_valid_filename(filename)
        fmtname = time.strftime(filename.encode("unicode-escape").decode()).encode().decode("unicode-escape")
        threading.Thread(target=asyncio.run, args=(self.danmaku_download_start(fmtname),)).start()
        if self.downloader == 'stream-gears':
            stream_gears_download(self.raw_stream_url, self.fake_headers, filename, config.get('segment_time'),
                                  config.get('file_size'))
        elif self.downloader == 'streamlink':
            parsed_url = urlparse(self.raw_stream_url)
            path = parsed_url.path
            if '.flv' in path: #streamlink无法处理flv,所以回退到ffmpeg
                self.ffmpeg_download(fmtname)
            else:
                self.streamlink_download(fmtname)
        else:
            self.ffmpeg_download(fmtname)

    def get_live_cover(self, uname, room_id, filename, timestamp, cover_url):
        import requests
        headers = self.fake_headers.copy()
        response = requests.get(cover_url, headers=headers, timeout=5)
        save_dir = f'cover/{uname}_{room_id}/'
        local_time = time.strftime('%Y-%m-%d_%H-%M-%S', time.localtime(timestamp))
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        cover_file_name = get_valid_filename(f'{filename}_{local_time}.png')
        live_cover_path = f'{save_dir}{cover_file_name}'
        if os.path.exists(live_cover_path):
            return live_cover_path
        else:
            with open(live_cover_path, 'wb') as f:
                f.write(response.content)
                return live_cover_path

    def streamlink_download(self, filename): #streamlink+ffmpeg混合下载模式，适用于下载hls流
        streamlink_input_args = ['--stream-segment-threads', '3', '--hls-playlist-reload-attempts', '1']
        streamlink_cmd = ['streamlink', *streamlink_input_args, self.raw_stream_url, 'best', '-O']
        ffmpeg_input_args = ['-reconnect_streamed', '1', '-reconnect_delay_max', '20', '-rw_timeout', '20000000']
        ffmpeg_cmd = ['ffmpeg', '-re', '-i', 'pipe:0', '-y',*ffmpeg_input_args, *self.default_output_args, *self.opt_args, '-c', 'copy', '-f', self.suffix]
        if config.get('segment_time'):
            ffmpeg_cmd += ['-f', 'segment',
                     f'{filename} part-%03d.{self.suffix}']
        else:
            ffmpeg_cmd += [
                f'{filename}.{self.suffix}.part']
        streamlink_proc = subprocess.Popen(streamlink_cmd, stdout=subprocess.PIPE)
        ffmpeg_proc = subprocess.Popen(ffmpeg_cmd, stdin=streamlink_proc.stdout, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        try:
            with ffmpeg_proc.stdout as stdout:
                for line in iter(stdout.readline, b''):
                    decode_line = line.decode(errors='ignore')
                    print(decode_line, end='', file=sys.stderr)
                    logger.debug(decode_line.rstrip())
            retval = ffmpeg_proc.wait()
        except KeyboardInterrupt:
            if sys.platform != 'win32':
                ffmpeg_proc.communicate(b'q')
            raise
        return retval

    def ffmpeg_download(self, filename):
        default_input_args = ['-headers', ''.join('%s: %s\r\n' % x for x in self.fake_headers.items()),
                              '-reconnect_streamed', '1', '-reconnect_delay_max', '20', '-rw_timeout', '20000000']
        parsed_url = urlparse(self.raw_stream_url)
        path = parsed_url.path
        if '.m3u8' in path:
            default_input_args += ['-max_reload', '3']
        args = ['ffmpeg', '-y', *default_input_args,
                '-i', self.raw_stream_url, *self.default_output_args, *self.opt_args,
                '-c', 'copy', '-f', self.suffix]
        if config.get('segment_time'):
            args += ['-f', 'segment',
                     f'{filename} part-%03d.{self.suffix}']
        else:
            args += [
                f'{filename}.{self.suffix}.part']

        proc = subprocess.Popen(args, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        try:
            with proc.stdout as stdout:
                for line in iter(stdout.readline, b''):  # b'\n'-separated lines
                    decode_line = line.decode(errors='ignore')
                    print(decode_line, end='', file=sys.stderr)
                    logger.debug(decode_line.rstrip())
            retval = proc.wait()
        except KeyboardInterrupt:
            if sys.platform != 'win32':
                proc.communicate(b'q')
            raise
        return retval

    async def danmaku_download_start(self, filename):
        pass

    def run(self):
        if not self.check_stream():
            return False
        file_name = self.file_name
        retval = self.download(file_name)
        logger.info(f'{retval}part: {file_name}.{self.suffix}')
        self.rename(f'{file_name}.{self.suffix}')
        return retval

    def start(self):
        i = 0
        logger.info('开始下载%s：%s' % (self.__class__.__name__, self.fname))
        date = time.localtime()
        lasttime_download = 0
        thistime_download = 0
        # delay = config.get('delay') if (config.get('delay') > 5) else 5
        delay = 30  # 有人的delay会设置的过长，遂写死为30
        while i < 30:
            try:
                thistime_download = time.time()
                ret = self.run()
            except:
                logger.exception("Uncaught exception:")
                continue
            finally:
                self.close()

            # 限制每次拉流的时间间隔必须大于delay，防止主播还没推流的时候疯狂重复请求
            interval = thistime_download - lasttime_download
            if interval < delay:
                logger.info(f"频繁请求：等待{interval}秒后再次请求直播流")
                time.sleep(delay - interval)
                lasttime_download = thistime_download
                continue

            if ret is False:
                if config.get('delay'):
                    time.sleep(config.get('delay'))
                    logger.info(f"delay: {config.get('delay')}")
                    if self.check_stream():
                        time.sleep(5)
                        continue
                break
            elif ret == 1 or self.downloader == 'stream-gears':
                time.sleep(45)


            i += 1
        logger.info(f'退出下载{i}: {self.fname}')
        return {
            'name': self.fname,
            'url': self.url,
            'title': self.room_title,
            'date': date,
            'live_cover_path': self.live_cover_path,
        }

    @staticmethod
    def rename(file_name):
        try:
            os.rename(file_name + '.part', file_name)
            logger.debug('更名{0}为{1}'.format(file_name + '.part', file_name))
        except FileNotFoundError:
            logger.info('FileNotFoundError:' + file_name)
        except FileExistsError:
            os.rename(file_name + '.part', file_name)
            logger.info('FileExistsError:更名{0}为{1}'.format(file_name + '.part', file_name))

    @property
    def file_name(self):
        if self.filename_prefix:  # 判断是否存在自定义录播命名设置
            filename = (self.filename_prefix.format(streamer=self.fname, title=self.room_title).encode(
                'unicode-escape').decode()).encode().decode("unicode-escape")
        else:
            filename = f'{self.fname}%Y-%m-%dT%H_%M_%S'
        filename = get_valid_filename(filename)
        return time.strftime(filename.encode("unicode-escape").decode()).encode().decode("unicode-escape")

    def close(self):
        pass


def stream_gears_download(url, headers, file_name, segment_time=None, file_size=None):
    class Segment:
        pass

    segment = Segment()
    if segment_time:
        seg_time = segment_time.split(':')
        print(int(seg_time[0]) * 60 * 60 + int(seg_time[1]) * 60 + int(seg_time[2]))
        segment.time = int(seg_time[0]) * 60 * 60 + int(seg_time[1]) * 60 + int(seg_time[2])
    if file_size:
        segment.size = file_size
    if file_size is None and segment_time is None:
        segment.size = 8 * 1024 * 1024 * 1024
    stream_gears.download(
        url,
        headers,
        file_name,
        segment
    )


def get_valid_filename(name):
    """
    Return the given string converted to a string that can be used for a clean
    filename. Remove leading and trailing spaces; convert other spaces to
    underscores; and remove anything that is not an alphanumeric, dash,
    underscore, or dot.
    # >>> get_valid_filename("john's portrait in 2004.jpg")
    >>> get_valid_filename("{self.fname}%Y-%m-%dT%H_%M_%S")
    '{self.fname}%Y-%m-%dT%H_%M_%S'
    """
    # s = str(name).strip().replace(" ", "_") #因为有些人会在主播名中间加入空格，为了避免和录播完毕自动改名冲突，所以注释掉
    s = re.sub(r"(?u)[^-\w.%{}\[\]【】「」\s]", "", str(name))
    if s in {"", ".", ".."}:
        raise RuntimeError("Could not derive file name from '%s'" % name)
    return s



import asyncio
import random
import subprocess
import threading
from urllib.parse import urlparse

import yt_dlp
import streamlink
import os
import time
from PIL import Image
from yt_dlp.utils import MaxDownloadsReached

from yt_dlp.utils import DateRange
from biliup.config import config
from biliup.plugins.Danmaku import DanmakuClient
from ..engine.decorators import Plugin
from ..engine.download import DownloadBase, get_valid_filename, stream_gears_download
from . import logger

VALID_URL_BASE = r'https?://(?:(?:www|m)\.)?youtube\.com/(?P<id>.*?)\??(.*?)'

@Plugin.download(regexp=VALID_URL_BASE)
class Youtube(DownloadBase):
    def __init__(self, fname, url, suffix='webm'):
        DownloadBase.__init__(self, fname, url, suffix=suffix)
        self.use_new_ytb_downloader = config.get('use_new_ytb_downloader', False)
        self.ytb_danmaku = config.get('ytb_danmaku', False)
        self.cookiejarFile = config.get('user', {}).get('youtube_cookie')
        self.vcodec = config.get('youtube_prefer_vcodec','av01|vp9|avc')
        self.acodec = config.get('youtube_prefer_acodec','opus|mp4a')
        self.resolution = config.get('youtube_max_resolution','4320')
        self.filesize = config.get('youtube_max_videosize','100G')
        self.beforedate = config.get('youtube_before_date','20770707')
        self.afterdate = config.get('youtube_after_date','19700101')
        self.use_youtube_cover = config.get('use_youtube_cover', False)

    def check_stream(self):
        if self.use_new_ytb_downloader and self.downloader == 'ffmpeg':
            _fname, self.room_title = self.get_stream_info(self.url)
            port = random.randint(1025, 65535)
            stream_shell = [
                "streamlink",
                "--player-external-http",  # 为外部程序提供流媒体数据
                "--player-external-http-port", str(port),  # 对外部输出流的端口
                self.url, "best"  # 流链接
            ]
            self.proc = subprocess.Popen(stream_shell)
            self.raw_stream_url = f"http://localhost:{port}"
            i = 0
            while i < 5:
                if not (self.proc.poll() is None):
                    return
                time.sleep(1)
                i += 1
            return True
        else:
            with yt_dlp.YoutubeDL({'download_archive': 'archive.txt', 'ignoreerrors': True, 'extract_flat': True,
                                   'cookiefile': self.cookiejarFile}) as ydl:
                info = ydl.extract_info(self.url, download=False, process=False)
                if info is None:
                    logger.warning(self.cookiejarFile)
                    logger.warning(self.url)
                    return False
                if '_type' in info:
                    # /live 形式链接取标题
                    if info['_type'] in 'url' and info['webpage_url_basename'] in 'live':
                        live = ydl.extract_info(info['url'], download=False, process=False)
                        self.room_title = live['title']
                    # Playlist 暂时返回列表名
                    if info['_type'] in 'playlist':
                        self.room_title = info['title']
                else:
                    # 视频取标题
                    self.room_title = info['title']
                if info.get('entries') is None:
                    if ydl.in_download_archive(info):
                        return False
                    return True
                for entry in info['entries']:
                    # 取 Playlist 内视频标题
                    self.room_title = entry['title']
                    if ydl.in_download_archive(entry):
                        continue
                    # ydl.record_download_archive()
                    return True

    def download(self, filename):
        if self.use_new_ytb_downloader and self.downloader == 'ffmpeg':
            if self.filename_prefix:  # 判断是否存在自定义录播命名设置
                filename = (self.filename_prefix.format(streamer=self.fname, title=self.room_title).encode(
                    'unicode-escape').decode()).encode().decode("unicode-escape")
            else:
                filename = f'{self.fname}%Y-%m-%dT%H_%M_%S'
            filename = get_valid_filename(filename)
            fmtname = time.strftime(filename.encode("unicode-escape").decode()).encode().decode("unicode-escape")
            threading.Thread(target=asyncio.run, args=(self.danmaku_download_start(fmtname),)).start()
            self.ffmpeg_download(fmtname)
        else:
            try:
                ydl_opts = {
                    'outtmpl': filename+ '.%(ext)s',
                    'format': f"bestvideo[vcodec~='^({self.vcodec})'][height<={self.resolution}][filesize<{self.filesize}]+bestaudio[acodec~='^({self.acodec})']/best[height<={self.resolution}]/best",
                    'cookiefile': self.cookiejarFile,
                    # 'proxy': proxyUrl,
                    'ignoreerrors': True,
                    'max_downloads': 1,
                    'daterange' : DateRange(self.afterdate, self.beforedate),
                    'break_on_reject': True,
                    'download_archive': 'archive.txt',
                }
                if self.use_youtube_cover is True:
                    ydl_opts['writethumbnail'] = True
                with yt_dlp.YoutubeDL(ydl_opts) as ydl:
                    ydl.download([self.url])
            except MaxDownloadsReached:
                save_dir = f'cover/youtube/'
                webp_cover_path = f'{filename}.webp'
                jpg_cover_path = f'{filename}.jpg'
                if os.path.exists(webp_cover_path):
                    with Image.open(webp_cover_path) as img:
                        img = img.convert('RGB')
                        if not os.path.exists(save_dir):
                            os.makedirs(save_dir)
                        img.save(f'{save_dir}{filename}.jpg', format='JPEG')
                    os.remove(webp_cover_path)
                    self.live_cover_path = f'{save_dir}{filename}.jpg'
                elif os.path.exists(jpg_cover_path):
                    os.rename(jpg_cover_path, f'{save_dir}{filename}.jpg')
                    self.live_cover_path = f'{save_dir}{filename}.jpg'
                return False
            except yt_dlp.utils.RejectedVideoReached:
                logger.info("已下载完毕指定日期内的视频，结束本次任务")
                return False
            except yt_dlp.utils.DownloadError:
                logger.exception(self.fname)
                return 1
            return 0

    def get_stream_info(self, url):
        session = streamlink.Streamlink()
        # Streamlink 插件系统的内部运行机制，通过 URL 匹配到 YouTube 插件
        plugin = session.resolve_url(url)[1]
        streams = plugin(session=session, url=url)
        streams._get_streams()
        author = streams.get_author()
        title = streams.get_title()
        return author, title


import logging
import os
import re
import shutil
import subprocess
from functools import reduce
from pathlib import Path

logger = logging.getLogger('biliup')


class UploadBase:
    def __init__(self, principal, data, persistence_path=None, postprocessor=None):
        self.principal = principal
        self.persistence_path = persistence_path
        self.data = data
        self.post_processor = postprocessor

    # @property
    @staticmethod
    def file_list(index):
        file_list = []
        for file_name in os.listdir('.'):
            if index in file_name and os.path.isfile(file_name):
                file_list.append(file_name)
        file_list = sorted(file_list)
        return file_list

    @staticmethod
    def remove_filelist(file_list):
        for r in file_list:
            os.remove(r)
            logger.info('删除-' + r)


    def filter_file(self, index):
        media_extensions = ['.mp4', '.flv', '.ts', '.flv.part']
        file_list = UploadBase.file_list(index)
        if len(file_list) == 0:
            return False
        for f in file_list:
            if f.endswith('.part'):
                new_name = os.path.splitext(f)[0]
                shutil.move(f, new_name)
                logger.info(f'{f}存在已更名为{new_name}')
        for r in file_list:
            name, ext = os.path.splitext(r)
            if ext in ('.mp4', '.flv', '.ts'):
                file_size = os.path.getsize(r) / 1024 / 1024
                threshold = self.data.get('threshold',2)
                if file_size <= threshold:
                    self.remove_file(r)
                    logger.info(f'过滤删除-{r}')
            if ext == '.xml': #过滤不存在对应视频的xml弹幕文件
                xml_file_name = name
                # media_regex = re.compile(r'^{}(\.(mp4|flv|ts))?$'.format(
                #     re.escape(xml_file_name)
                # ))
                # if not any(media_regex.match(f'{xml_file_name}{ext2}') for ext2 in media_extensions for x in file_list):
                #     self.remove_file(r)
                #     logger.info(f'无视频，已过滤删除-{r}')
                have_video = False
                for video_ext in media_extensions:
                    if f"{xml_file_name}{video_ext}" in file_list: have_video = True
                if not have_video:
                    self.remove_file(r)
                    logger.info(f'无视频，已过滤删除-{r}')
        file_list = UploadBase.file_list(index)
        if len(file_list) == 0:
            logger.info('视频过滤后无文件可传')
            return False



        return True

    def remove_file(self, file_path):
        os.remove(file_path)

    def upload(self, file_list):
        raise NotImplementedError()

    def start(self):
        if self.filter_file(self.principal):
            logger.info('准备上传' + self.data["format_title"])
            needed2process = self.upload(UploadBase.file_list(self.principal))
            if needed2process:
                self.postprocessor(needed2process)

    def postprocessor(self, data):
        # data = file_list
        if self.post_processor is None:
            return self.remove_filelist(data)
        for post_processor in self.post_processor:
            if post_processor == 'rm':
                self.remove_filelist(data)
                continue
            if post_processor.get('mv'):
                for file in data:
                    path = Path(file)
                    dest = Path(post_processor['mv'])
                    if not dest.is_dir():
                        dest.mkdir(parents=True, exist_ok=True)
                    try:
                        shutil.move(path, dest / path.name)
                    except Exception as e:
                        logger.exception(e)
                        continue
                    logger.info(f"move to {(dest / path.name).absolute()}")
            if post_processor.get('run'):
                try:
                    process_output = subprocess.check_output(
                        post_processor['run'], shell=True, 
                        input=reduce(lambda x, y: x + str(Path(y).absolute()) + '\n', data, ''),
                        stderr=subprocess.STDOUT, text=True)
                    logger.info(process_output.rstrip())
                except subprocess.CalledProcessError as e:
                    logger.exception(e.output)
                    continue



# 部分弹幕功能代码来自项目：https://github.com/IsoaSFlus/danmaku，感谢大佬
# 快手弹幕代码来源及思路：https://github.com/py-wuhao/ks_barrage，感谢大佬
# 部分斗鱼录播修复代码与思路来源于：https://github.com/SmallPeaches/DanmakuRender，感谢大佬
# 仅抓取用户弹幕，不包括入场提醒、礼物赠送等。

import asyncio
import os
import re
import ssl
import time
import logging
from functools import partial

import lxml.etree as etree
import aiofiles
import aiohttp

from biliup.plugins.Danmaku.douyu import Douyu
from biliup.plugins.Danmaku.huya import Huya
from biliup.plugins.Danmaku.bilibili import Bilibili
from biliup.plugins.Danmaku.twitch import Twitch

logger = logging.getLogger('biliup')
__all__ = ['DanmakuClient']


class DanmakuClient:
    def __init__(self, url, filename):
        self.__starttime = time.time()
        self.__filename = os.path.splitext(filename)[0] + '.xml'
        self.__filename_video_suffix = filename
        self.__url = ''
        self.__site = None
        self.__hs = None
        self.__ws = None
        self.__stop = False
        self.__dm_queue = None
        self.__link_status = True

        if 'http://' == url[:7] or 'https://' == url[:8]:
            self.__url = url
        else:
            self.__url = 'http://' + url
        for u, s in {'douyu.com': Douyu,
                     'huya.com': Huya,
                     'live.bilibili.com': Bilibili,
                     'twitch.tv': Twitch
                     }.items():
            if re.match(r'^(?:http[s]?://)?.*?%s/(.+?)$' % u, url):
                self.__site = s
                self.__u = u
                break
        if self.__site is None:
            print('Invalid link!')
            exit()

    async def init_ws(self):
        ws_url, reg_datas = await self.__site.get_ws_info(self.__url)
        ctx = ssl.create_default_context()
        ctx.set_ciphers('DEFAULT')
        try:
            self.__ws = await self.__hs.ws_connect(ws_url, ssl_context=ctx)
            for reg_data in reg_datas:
                if type(reg_data) == str:
                    await self.__ws.send_str(reg_data)
                else:
                    await self.__ws.send_bytes(reg_data)
        except Exception as Error:
            logger.debug(f"init_ws：捕获到异常：{Error}")
        if not os.path.exists(self.__filename):
            async with aiofiles.open(self.__filename, mode='w') as f:
                await f.write("<?xml version='1.0' encoding='UTF-8'?>\n"
                            "<i xmlns:ns0='http://www.w3.org/1999/XSL/Transform'>\n"
                            "</i>"
                            )

    async def heartbeats(self):
        while not self.__stop and self.__site.heartbeat:
            await asyncio.sleep(self.__site.heartbeatInterval)
            try:
                if type(self.__site.heartbeat) == str:
                    await self.__ws.send_str(self.__site.heartbeat)
                else:
                    await self.__ws.send_bytes(self.__site.heartbeat)
            except Exception as Error:
                logger.debug(f"heartbeats：捕获到异常：{Error}")
                if not self.__stop:
                    logger.info(f'触发弹幕重连')
                    await self.init_ws()
                    await asyncio.sleep(3)
                
        await self.__hs.close()

    async def fetch_danmaku(self):
        # while not self.__stop:
        await asyncio.sleep(1)
        await self.init_ws()
        await asyncio.sleep(1)
        async for msg in self.__ws:
            try:
                if self.__stop:
                    await self.__dm_queue.put(None)
                    return
                # self.__link_status = True
                ms = self.__site.decode_msg(msg.data)
                for m in ms:
                    await self.__dm_queue.put(m)
            except Exception as Error:
                logger.debug(f"fetch_danmaku：弹幕处理异常：{Error}")
                await asyncio.sleep(10)

    async def print_danmaku(self):
        parser = etree.XMLParser(recover=True)
        tree = etree.parse(self.__filename, parser=parser)
        root = tree.getroot()
        msg_i = 0
        msg_col = {'0': '16777215', '1': '16717077', '2': '2000880', '3': '8046667', '4': '16744192', '5': '10172916',
                   '6': '16738740'}

        def write_file(filename):
            with open(filename, "wb") as f:
                etree.indent(root, "\t")
                tree.write(f, encoding="UTF-8", xml_declaration=True, pretty_print=True)

        while not self.__stop:
            try:
                m = await self.__dm_queue.get()
                if m is None:
                    return
                if m['msg_type'] == 'danmaku':
                    d = etree.SubElement(root, 'd')
                    if 'col' in m:
                        color = msg_col[m["col"]]
                    elif 'color' in m:
                        color = m["color"]
                    else:
                        color = '16777215'
                    msg_time = format(time.time() - self.__starttime, '.3f')
                    d.set('p', f"{msg_time},1,25,{color},0,0,0,0")
                    d.text = m["content"]
            except Exception as Error:
                logger.debug(f"print_danmaku：弹幕处理异常：{Error}")
                
            try:
                if m['msg_type'] == 'danmaku' and msg_i >= 5:
                    # loop = asyncio.get_running_loop()
                    # await loop.run_in_executor(None, partial(write_file, self.__filename))
                    write_file(self.__filename)
                    msg_i = 0
                else:
                    msg_i = msg_i + 1
            except Exception as Error:
                logger.debug(f"print_danmaku：弹幕写入异常：{Error}")


    async def start(self):
        self.__dm_queue = asyncio.Queue()
        self.__hs = aiohttp.ClientSession()
        await self.init_ws()
        await asyncio.gather(
            self.heartbeats(),
            self.fetch_danmaku(),
            self.print_danmaku(),
        )

    def stop(self):
        self.__stop = True
        self.__hs.close()
        if not (os.path.exists(f"{self.__filename_video_suffix}.part") or
                os.path.exists(f"{self.__filename_video_suffix}")):
            os.remove(self.__filename)



# 虎牙直播：https://www.huya.com/lpl
# 斗鱼直播：https://www.douyu.com/9999


import re

import aiohttp

from .tars import tarscore


class Huya:
    wss_url = 'wss://cdnws.api.huya.com/'
    heartbeat = b'\x00\x03\x1d\x00\x00\x69\x00\x00\x00\x69\x10\x03\x2c\x3c\x4c\x56\x08\x6f\x6e\x6c\x69\x6e\x65\x75' \
                b'\x69\x66\x0f\x4f\x6e\x55\x73\x65\x72\x48\x65\x61\x72\x74\x42\x65\x61\x74\x7d\x00\x00\x3c\x08\x00' \
                b'\x01\x06\x04\x74\x52\x65\x71\x1d\x00\x00\x2f\x0a\x0a\x0c\x16\x00\x26\x00\x36\x07\x61\x64\x72\x5f' \
                b'\x77\x61\x70\x46\x00\x0b\x12\x03\xae\xf0\x0f\x22\x03\xae\xf0\x0f\x3c\x42\x6d\x52\x02\x60\x5c\x60' \
                b'\x01\x7c\x82\x00\x0b\xb0\x1f\x9c\xac\x0b\x8c\x98\x0c\xa8\x0c '
    heartbeatInterval = 60

    @staticmethod
    async def get_ws_info(url):
        try:
            reg_datas = []
            url = 'https://m.huya.com/' + url.split('/')[-1]
            headers = {
                'user-agent': 'Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, '
                            'like Gecko) Chrome/79.0.3945.88 Mobile Safari/537.36'}
            async with aiohttp.ClientSession() as session:
                async with session.get(url, headers=headers) as resp:
                    room_page = await resp.text()
                    m = re.search(r"lYyid\":([0-9]+)", room_page, re.MULTILINE)
                    ayyuid = m.group(1)
                    m = re.search(r"lChannelId\":([0-9]+)", room_page, re.MULTILINE)
                    tid = m.group(1)
                    m = re.search(r"lSubChannelId\":([0-9]+)", room_page, re.MULTILINE)
                    sid = m.group(1)

            oos = tarscore.TarsOutputStream()
            oos.write(tarscore.int64, 0, int(ayyuid))
            oos.write(tarscore.boolean, 1, True)  # Anonymous
            oos.write(tarscore.string, 2, "")  # sGuid
            oos.write(tarscore.string, 3, "")
            oos.write(tarscore.int64, 4, int(tid))
            oos.write(tarscore.int64, 5, int(sid))
            oos.write(tarscore.int64, 6, 0)
            oos.write(tarscore.int64, 7, 0)

            wscmd = tarscore.TarsOutputStream()
            wscmd.write(tarscore.int32, 0, 1)
            wscmd.write(tarscore.bytes, 1, oos.getBuffer())

            reg_datas.append(wscmd.getBuffer())
        except:
            print("huya：get_ws_info()执行错误")
        finally:
            return Huya.wss_url, reg_datas

    @staticmethod
    def decode_msg(data):
        class user(tarscore.struct):
            def readFrom(ios):
                return ios.read(tarscore.string, 2, False).decode("utf8")

        class dcolor(tarscore.struct):
            def readFrom(ios):
                return ios.read(tarscore.int32, 0, False)

        name = ""
        content = ""
        msgs = []
        ios = tarscore.TarsInputStream(data)

        if ios.read(tarscore.int32, 0, False) == 7:
            ios = tarscore.TarsInputStream(ios.read(tarscore.bytes, 1, False))
            if ios.read(tarscore.int64, 1, False) == 1400:
                ios = tarscore.TarsInputStream(ios.read(tarscore.bytes, 2, False))
                name = ios.read(user, 0, False)  # username
                content = ios.read(tarscore.string, 3, False).decode("utf8")  # content
                color = ios.read(dcolor, 6, False)  # danmaku color
                if color == -1:
                    color = 16777215

        if name != "":
            msg = {"name": name, "color": f"{color}", "content": content, "msg_type": "danmaku"}
        # else:
        #     msg = {"name": "", "content": "", "msg_type": "other"}
            msgs.append(msg)
        return msgs
